#include "f12.h"
#include "dco.hpp"
#include <random>

template <typename T, typename Tp, typename To>
void adjoint_chunk(typename dco::mode<T>::external_adjoint_object_t *EA) {
  // local passive inputs to target code
  int ns=EA->template read_data<int>(); int nc=EA->template read_data<int>();
  int jc=EA->template read_data<int>(); To dt=EA->template read_data<To>();
  // read-only active input to target code
  T x_c=EA->template read_data<T>(); 
  dco::mode<T>::global_tape->register_variable(x_c);
  // read-only global active inputs to target code
  std::vector<std::vector<Tp>>* p_p=EA->template read_data<std::vector<std::vector<Tp>>*>(); 
  // mark start of target code
  typename dco::mode<T>::tape_t::position_t pos=dco::mode<T>::global_tape->get_position();
  T x=x_c;
  for (int j=jc;j<jc+ns/nc;++j) {
    To Z=EA->template read_data<To>();
    x+=dt*f1(x,(*p_p)[0][j])+f2(x,(*p_p)[1][j])*sqrt(dt)*Z;
  }
  // seed local active output of target code using the
  // stored reference to the adjoint of its twin
  dco::derivative(x)=EA->get_output_adjoint();
  // interpret tape of target code 
  // increments adjoints of global read-only *p_p implicitely
  dco::mode<T>::global_tape->interpret_adjoint_and_reset_to(pos);
  // increment adjoint of local active input explicitely
  EA->increment_input_adjoint(dco::derivative(x_c));
}

template <typename T, typename Tp, typename To=float>
T f(int np, int nb, int ns, int nc, T x, const std::vector<std::vector<Tp>>& p) {
  using namespace std;
  default_random_engine g; normal_distribution<To> d(0,1);
  T y=0, x_c=x; To dt=1./ns;
  for (int ib=0;ib<np;ib+=np/nb) 
    for (int i=ib;i<ib+np/nb;++i) {
      x=x_c;
      for (int jc=0;jc<ns;jc+=ns/nc) {
        // target individual chunks
        typename dco::mode<T>::external_adjoint_object_t* EA=dco::mode<T>::global_tape->template create_callback_object<typename dco::mode<T>::external_adjoint_object_t>();
        EA->write_data(ns); EA->write_data(nc); EA->write_data(jc);
        EA->write_data(dt); EA->write_data(x); EA->write_data(&p);
        // store reference to adjoint of non-constant local active 
        // input to be incremented with the adjoint of its twin
        EA->register_input(x);
        // execute passive target code 
        dco::mode<T>::global_tape->switch_to_passive();
        for (int j=jc;j<jc+ns/nc;++j) {
          To Z=d(g); EA->write_data(Z);
          x+=dt*f1(x,p[0][j])+f2(x,p[1][j])*sqrt(dt)*Z;
        }
        dco::mode<T>::global_tape->switch_to_active();
        // store reference to adjoint of local active output to
        // be used to seed its twin
        x=EA->register_output(dco::value(x));
        dco::mode<T>::global_tape->template insert_callback<typename dco::mode<T>::external_adjoint_object_t>(adjoint_chunk<T,Tp,To>,EA);
      }
      y+=x;
  }
  return pow(y/np,2);
}

int main(int argc, char* argv[]) {
  assert(argc==5);
  int np=std::stoi(argv[1]), nb=std::stoi(argv[2]);
  assert(np>0&&nb>0&&np%nb==0);
  int ns=std::stoi(argv[3]), nc=std::stoi(argv[4]);
  assert(ns>0&&nc>0&&ns%nc==0);
  using Tbb=double; std::cout.precision(std::numeric_limits<Tbb>::digits10);
  using Tb=dco::gt1s<Tbb>::type; using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x=1,y; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create(); 
  dco::ga1s<Tb>::global_tape->register_variable(x); 
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  dco::ga1s<Tb>::tape_t::position_t pos=dco::ga1s<Tb>::global_tape->get_position();
  // d(dy/d(x,p))dx
  dco::derivative(dco::value(x))=1; 
  y=f(np,nb,ns,nc,x,p); 
  dco::value(dco::derivative(y))=1; 
  dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
  std::cout << "d(dy/dx)/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
  for (int j=0;j<ns;++j) {
    std::cout << "d(dy/dp[0][" << j << "])/dx=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
    dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
    std::cout << "d(dy/dp[1][" << j << "])/dx=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
    dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
  }
  dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
  // d(dy/d(x,p))dp
  for (int i=0;i<ns;++i) {
    dco::derivative(dco::value(p[0][i]))=1;
    y=f(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[0][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[0][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[0][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[0][i]))=0;
    dco::derivative(dco::value(p[1][i]))=1;
    y=f(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[1][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[1][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[1][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[1][i]))=0;
  }
  std::cout << "y=" << dco::value(dco::value(y)) << std::endl;
  return 0;
}

